class SwitchState:
	"""
	Class to represent the switch state of a Crownstone.

	intensity:     Combined value as percentage: 0 - 100.
	relay:         Either on (True) or off (False).
	dimmer:        Dimmer percentage: 0 - 100.
	raw:           Raw value.
	"""
	def __init__(self, rawSwitchState: int = 0):
		"""
		:param rawSwitchState: state according to protocol (uint8).
		"""
		self.relay = (rawSwitchState & 0x80 == 1)

		self.dimmer = (rawSwitchState & 0x7F)
		if (self.dimmer > 100):
			self.dimmer = 100

		self.intensity = rawSwitchState
		if self.intensity > 100:
			self.intensity = 100

		self.raw = rawSwitchState

	def toString(self):
		return f"SwitchState(intensity={self.intensity}, relay={self.relay}, dimmer={self.dimmer})"

	def __str__(self):
		return self.toString()



import time

from crownstone_core.util.Conversion import Conversion
from crownstone_core.util.Timestamp import reconstructTimestamp


def parseOpCode3_type1(serviceData, data):
    if len(data) == 16:
        # dataType = data[0]
        
        serviceData.errorMode = True
        
        serviceData.crownstoneId = data[1]
        serviceData.errorsBitmask = Conversion.uint8_array_to_uint32([
            data[2],
            data[3],
            data[4],
            data[5]
        ])
        
        serviceData.errorTimestamp = Conversion.uint8_array_to_uint32([
            data[6],
            data[7],
            data[8],
            data[9]
        ])
        
        serviceData.flagsBitmask = data[10]
        # bitmask states
        bitmaskArray = Conversion.uint8_to_bit_array(serviceData.flagsBitmask)
        
        serviceData.dimmerReady   = bitmaskArray[0]
        serviceData.dimmingAllowed     = bitmaskArray[1]
        serviceData.hasError           = bitmaskArray[2]
        serviceData.switchLocked       = bitmaskArray[3]
        serviceData.timeIsSet          = bitmaskArray[4]
        serviceData.switchCraftEnabled = bitmaskArray[5]
        
        serviceData.temperature = Conversion.uint8_to_int8(data[11])

        serviceData.partialTimestamp = Conversion.uint8_array_to_uint16([data[12], data[13]])
        serviceData.uniqueIdentifier = serviceData.partialTimestamp

        if serviceData.timeIsSet:
            serviceData.timestamp = reconstructTimestamp(time.time(), serviceData.partialTimestamp)
        else:
            serviceData.timestamp = serviceData.partialTimestamp # this is now a counter
        
        realPower = Conversion.uint16_to_int16(
            Conversion.uint8_array_to_uint16([
                data[14],
                data[15]
            ])
        )
        
        serviceData.powerUsageReal = float(realPower) / 8.0
        
        # this packet has no validation
        serviceData.validation = 0



from crownstone_core.packets.serviceDataParsers.opCode3.opCode3_type1 import parseOpCode3_type1


def parseOpCode3_type3(serviceData, data):
    if len(data) == 16:
        parseOpCode3_type1(serviceData, data)

        # apply differences between type 1 and type 4
        serviceData.stateOfExternalCrownstone = True
        serviceData.powerUsageReal = 0
        serviceData.validation = data[15]



import uuid

class EventBus:

    def __init__(self):
        self.topics = {}
        self.subscriberIds = {}

    def subscribe(self, topic: object, callback: object) -> object:
        if topic not in self.topics:
            self.topics[topic] = {}

        subscriptionId = str(uuid.uuid4())
        self.subscriberIds[subscriptionId] = topic
        self.topics[topic][subscriptionId] = callback

        return subscriptionId

    def emit(self, topic, data = True):
        if topic in self.topics:
            callbackIds = list(self.topics[topic].keys())
            for subscriptionId in callbackIds:
                self.topics[topic][subscriptionId](data)


    def unsubscribe(self, subscriptionId):
        if subscriptionId is None:
            return

        if subscriptionId in self.subscriberIds:
            topic = self.subscriberIds[subscriptionId]
            if topic in self.topics:
                self.topics[topic].pop(subscriptionId)

            self.subscriberIds.pop(subscriptionId)
        else:
            pass
            #print("ERROR: BluenetEventBus: Subscription ID ", subscriptionId, " cannot be found.")




from crownstone_core.util.DataStepper import DataStepper
from crownstone_core.protocol.BluenetTypes import PowerSamplesType
from crownstone_core.Exceptions import CrownstoneError, CrownstoneException

class PowerSamplesPacket:
	def __init__(self, data):
		self.samplesType = PowerSamplesType.UNSPECIFIED
		self.index = 0            # uint8
		self.count = 0            # uint16
		self.timestamp = 0        # uint32
		self.delayUs = 0          # uint16
		self.sampleIntervalUs = 0 # uint16
		self.reserved = 0         # 2 bytes
		self.offset = 0           # int16
		self.multiplier = 0.0     # float

		self.samples = []         # int16 list
		self.load(data)

	def load(self, data):
		"""
		Parses data buffer to set member variables.

		data : list of bytes

		Raises exception when parsing fails.
		"""
		streamBuf = DataStepper(data)

		samplesTypeVal = streamBuf.getUInt8()
		self.samplesType = PowerSamplesType(samplesTypeVal) # Throws exception of value is not in enum
		self.index = streamBuf.getUInt8()
		self.count = streamBuf.getUInt16()
		self.timestamp = streamBuf.getUInt32()
		self.delayUs = streamBuf.getUInt16()
		self.sampleIntervalUs = streamBuf.getUInt16()
		streamBuf.skip(2)
		self.offset = streamBuf.getInt16()
		self.multiplier = streamBuf.getFloat()
		self.samples = []
		for i in range(0, self.count):
			self.samples.append(streamBuf.getInt16())

	def toString(self):
		msg = "PowerSamplesPacket("
		msg += "type=" + str(self.samplesType)
		msg += " count=" + str(self.count)
		msg += " timestamp=" + str(self.timestamp)
		msg += " delayUs=" + str(self.delayUs)
		msg += " sampleIntervalUs=" + str(self.sampleIntervalUs)
		msg += " offset=" + str(self.offset)
		msg += " multiplier=" + str(self.multiplier)
		msg += " samples=" + str(self.samples)
		msg += ")"
		return msg

	def __str__(self):
		return self.toString()


import time

from crownstone_core.protocol.SwitchState import SwitchState
from crownstone_core.util.Conversion import Conversion
from crownstone_core.util.DataStepper import DataStepper
from crownstone_core.util.Timestamp import reconstructTimestamp


def parseOpCode7_type4(serviceData, data):
    if len(data) == 16:
        # dataType = data[0]
        
        serviceData.stateOfExternalCrownstone = False

        payload = DataStepper(data)

        payload.skip() # first byte is the  datatype.
        serviceData.crownstoneId = payload.getUInt8()
        serviceData.switchState = SwitchState(payload.getUInt8())
        serviceData.flagsBitmask = payload.getUInt8()
        serviceData.behaviourMasterHash = payload.getUInt16()
        payload.skip(6)
        serviceData.partialTimestamp = payload.getUInt16()
        payload.skip()
        serviceData.validation = payload.getUInt8()

        # bitmask states
        bitmaskArray = Conversion.uint8_to_bit_array(serviceData.flagsBitmask)
        
        serviceData.dimmerReady        = bitmaskArray[0]
        serviceData.dimmingAllowed     = bitmaskArray[1]
        serviceData.hasError           = bitmaskArray[2]
        serviceData.switchLocked       = bitmaskArray[3]
        serviceData.timeIsSet          = bitmaskArray[4]
        serviceData.switchCraftEnabled = bitmaskArray[5]

        serviceData.tapToToggleEnabled = bitmaskArray[6]
        serviceData.behaviourOverridden = bitmaskArray[7]

        serviceData.partialTimestamp = Conversion.uint8_array_to_uint16([data[12], data[13]])
        serviceData.uniqueIdentifier = serviceData.partialTimestamp
        
        if serviceData.timeIsSet:
            serviceData.timestamp = reconstructTimestamp(time.time(), serviceData.partialTimestamp)
        else:
            serviceData.timestamp = serviceData.partialTimestamp # this is now a counter



from crownstone_core.util.Conversion import Conversion


class CrownstoneErrors:
    
    def __init__(self, bitMask):
        self.bitMask = bitMask
    
        bitArray = Conversion.uint32_to_bit_array_reversed(bitMask)
    
        self.overCurrent        = bitArray[31 - 0]
        self.overCurrentDimmer  = bitArray[31 - 1]
        self.temperatureChip    = bitArray[31 - 2]
        self.temperatureDimmer  = bitArray[31 - 3]
        self.dimmerOnFailure    = bitArray[31 - 4]
        self.dimmerOffFailure   = bitArray[31 - 5]
    
    def hasErrors(self):
        return self.bitMask == 0
    
    def getDictionary(self):
        returnDict = {
            "overCurrent":        self.overCurrent,
            "overCurrentDimmer":  self.overCurrentDimmer,
            "temperatureChip":    self.temperatureChip,
            "temperatureDimmer":  self.temperatureDimmer,
            "dimmerOnFailure":    self.dimmerOnFailure,
            "dimmerOffFailure":   self.dimmerOffFailure,
            "bitMask":            self.bitMask
        }
        
        return returnDict
    


from crownstone_core.protocol.SwitchState import SwitchState
from crownstone_core.util.Conversion import Conversion

def parseOpCode4_type0(serviceData, data):
    if len(data) == 16:
        # dataType = data[0]

        serviceData.switchState  = SwitchState(data[1])
        serviceData.flagsBitmask = data[2]
        
        # bitmask states
        bitmaskArray                 = Conversion.uint8_to_bit_array(serviceData.flagsBitmask)
        serviceData.dimmerReady      = bitmaskArray[0]
        serviceData.dimmingAllowed   = bitmaskArray[1]
        serviceData.hasError         = bitmaskArray[2]
        serviceData.switchLocked     = bitmaskArray[3]
        serviceData.timeIsSet        = bitmaskArray[4]
        
        serviceData.temperature  = Conversion.uint8_to_int8(data[3])
        
        powerFactor = Conversion.uint8_to_int8(data[4])
        realPower = Conversion.uint16_to_int16(Conversion.uint8_array_to_uint16([data[5], data[6]]))

        serviceData.powerFactor = float(powerFactor) / 127.0
        
        # we cannot have a 0 for a power factor.To avoid division by 0, we set it to be either 0.01 or -0.01
        if 0 <= serviceData.powerFactor < 0.01:
            serviceData.powerFactor = 0.01
        elif -0.01 < serviceData.powerFactor < 0:
            serviceData.powerFactor = -0.01

        
        serviceData.powerUsageReal = float(realPower) / 8.0
        serviceData.powerUsageApparent = serviceData.powerUsageReal / serviceData.powerFactor
        
        serviceData.accumulatedEnergy = Conversion.uint32_to_int32(
            Conversion.uint8_array_to_uint32([
                data[7],
                data[8],
                data[9],
                data[10]
            ])
        )
        serviceData.accumulatedEnergy = serviceData.accumulatedEnergy * 64 # correction to display in J

        serviceData.uniqueIdentifier = data[11]


class DeviceCharacteristics:
    HardwareRevision = "2a27"
    FirmwareRevision = "2a26"


class CrownstoneCharacteristics:
    SessionData    = "24f0000e-7d10-4805-bfc1-7663a01c3bff"
    FactoryReset   = "24f00009-7d10-4805-bfc1-7663a01c3bff"
    Control        = "24f0000c-7d10-4805-bfc1-7663a01c3bff"
    Result         = "24f0000d-7d10-4805-bfc1-7663a01c3bff"


class SetupCharacteristics:
    MacAddress     = "24f10002-7d10-4805-bfc1-7663a01c3bff"
    SessionKey     = "24f10003-7d10-4805-bfc1-7663a01c3bff"
    GoToDFU        = "24f10006-7d10-4805-bfc1-7663a01c3bff"
    SessionData    = "24f1000e-7d10-4805-bfc1-7663a01c3bff"
    SetupControl   = "24f1000c-7d10-4805-bfc1-7663a01c3bff"
    Result         = "24f1000d-7d10-4805-bfc1-7663a01c3bff"

class DFUCharacteristics:
    ControlPoint   = "00001531-1212-EFDE-1523-785FEABCD123"
    Packet         = "00001532-1212-EFDE-1523-785FEABCD123"




from crownstone_core.packets.ServiceData import ServiceData
from crownstone_core.protocol.Services import DFU_ADVERTISEMENT_SERVICE_UUID
from crownstone_core.util.Conversion import Conversion


class Advertisement:
    
    def __init__(self, address, rssi, nameText, serviceData, serviceUUID=None):
        self.address = address
        self.rssi = rssi
        self.name = nameText

        self.serviceUUID = serviceUUID
        self.serviceData = None
        self.operationMode = None

        dataString = serviceData
        
        if serviceData is not None:
            if isinstance(serviceData, str):
                dataArray = Conversion.hex_string_to_uint8_array(dataString)
                self.serviceUUID = Conversion.uint8_array_to_uint16([dataArray[0], dataArray[1]])
                # pop the service UUID
                dataArray.pop(0)
                dataArray.pop(0)
            else:
                dataArray = serviceData

            if dataArray:
                self.serviceData = ServiceData(dataArray)
    
            self.operationMode = "NORMAL"

    
    def isInDFUMode(self):
        return False
    
    def isInSetupMode(self):
        if self.serviceData is not None:
            return self.serviceData.setupMode
        return False
    
    def isCrownstoneFamily(self):
        return self.serviceUUID == 0xC001 or self.serviceUUID == 0xC002 or self.serviceUUID == 0xC003 or self.serviceUUID == DFU_ADVERTISEMENT_SERVICE_UUID

    def hasScanResponse(self):
        return self.serviceData is not None
    
    def getCrownstoneId(self):
        if self.hasScanResponse() and self.isCrownstoneFamily():
            return self.serviceData.crownstoneId
    
    def decrypt(self, key):
        if self.serviceData:
            self.serviceData.decrypt(key)
            
    def getDictionary(self):
        data = {}
    
        data["name"] = self.name
        data["rssi"] = self.rssi
        data["address"] = self.address
        
        if self.serviceUUID is not None:
            data["serviceUUID"] = self.serviceUUID
            
        if self.serviceData is not None:
            data["serviceData"] = self.serviceData.getDictionary()
    
        return data
    
    def getSummary(self):
        data = {}
        if self.serviceData is not None:
            data = self.serviceData.getSummary(self.address)
        return data

    



from enum import IntEnum

DAY_START_TIME_SECONDS_SINCE_MIDNIGHT = 4*3600

class BehaviourType(IntEnum):
    behaviour  = 0
    twilight   = 1
    smartTimer = 2

    @classmethod
    def has_value(cls, value):
        return any(value == item.value for item in cls)




import time

from crownstone_core.protocol.SwitchState import SwitchState
from crownstone_core.util.Conversion import Conversion
from crownstone_core.util.Timestamp import reconstructTimestamp


def parseOpCode3_type0(serviceData, data):
    if len(data) == 16:
        # dataType = data[0]
        
        serviceData.stateOfExternalCrownstone = False
        
        serviceData.crownstoneId = data[1]
        serviceData.switchState  = SwitchState(data[2])
        serviceData.flagsBitmask = data[3]

        # bitmask states
        bitmaskArray = Conversion.uint8_to_bit_array(serviceData.flagsBitmask)
        
        serviceData.dimmerReady        = bitmaskArray[0]
        serviceData.dimmingAllowed     = bitmaskArray[1]
        serviceData.hasError           = bitmaskArray[2]
        serviceData.switchLocked       = bitmaskArray[3]
        serviceData.timeIsSet          = bitmaskArray[4]
        serviceData.switchCraftEnabled = bitmaskArray[5]

        serviceData.tapToToggleEnabled = bitmaskArray[6]
        serviceData.behaviourOverridden = bitmaskArray[7]

        serviceData.temperature        = Conversion.uint8_to_int8(data[4])
        powerFactor                    = Conversion.uint8_to_int8(data[5])
        realPower                      = Conversion.uint16_to_int16(Conversion.uint8_array_to_uint16([data[6], data[7]]))
        
        serviceData.powerFactor = float(powerFactor) / 127.0
        
        # we cannot have a 0 for a power factor.To avoid division by 0, we set it to be either 0.01 or -0.01
        if 0 <= serviceData.powerFactor < 0.01:
            serviceData.powerFactor = 0.01
        elif -0.01 < serviceData.powerFactor < 0:
            serviceData.powerFactor = -0.01
        
        serviceData.powerUsageReal = float(realPower) / 8.0
        serviceData.powerUsageApparent = serviceData.powerUsageReal / serviceData.powerFactor
        
        serviceData.accumulatedEnergy = Conversion.uint32_to_int32(
            Conversion.uint8_array_to_uint32([
                data[8],
                data[9],
                data[10],
                data[11]
            ])
        )
        serviceData.accumulatedEnergy = serviceData.accumulatedEnergy * 64  # correction to display in J
        
        serviceData.partialTimestamp = Conversion.uint8_array_to_uint16([data[12], data[13]])
        serviceData.uniqueIdentifier = serviceData.partialTimestamp
        
        if serviceData.timeIsSet:
            serviceData.timestamp = reconstructTimestamp(time.time(), serviceData.partialTimestamp)
        else:
            serviceData.timestamp = serviceData.partialTimestamp # this is now a counter

        globalBitmaskArray = Conversion.uint8_to_bit_array(data[14])
        serviceData.behaviourEnabled = globalBitmaskArray[0]
        
        serviceData.validation = data[15]

    




from crownstone_core.protocol.BluenetTypes import ControlType
from crownstone_core.util.DataStepper import DataStepper


class SessionDataPacket:
    
    def __init__(self, data) :
        self.validation = ControlType.UNSPECIFIED
        self.protocol = 0
        self.sessionNonce = 0
        self.validationKey = 0
        self.valid = True

        self.payload = []
        self.load(data)
    

    def load(self, data):
        if len(data) == 16:
            payload = DataStepper(data)

            self.validation = payload.getUInt32()
            self.protocol = payload.getUInt8()
            self.sessionNonce = payload.getAmountOfBytes(5)
            self.validationKey = payload.getAmountOfBytes(4)
        else:
            self.valid = False



# BLE specification's rules are defined in Core spec 4.2 Vol. 6 Part B Chapter 2.1.2 and are as follows:

# The initiator shall ensure that the Access Address meets the following requirements:
# • It shall not be the advertising channel packets’ Access Address.
# • It shall not have all four octets equal.
# • It shall not be a sequence that differs from the advertising channel packets’ Access Address by only one bit.
# • It shall have no more than six consecutive zeros or ones.
# • It shall have no more than 24 transitions.
# • It shall have a minimum of two transitions in the most significant six bits.
import random

from crownstone_core.util.Conversion import Conversion

ADVERTISING_ACCESS_ADDRESS = 0x8E89BED6

class Util:
    
    @staticmethod
    def generateMeshAccessAddress(retries = 0):
        randomArray = [random.randint(0,255), random.randint(0,255), random.randint(0,255), random.randint(0,255)]
        if not Util.validateMeshAccessAddress(randomArray) and retries < 100:
            retries += 1
            return Util.generateMeshAccessAddress(retries)
        
        return Conversion.uint8_array_to_uint32(randomArray)

    @staticmethod
    def validateMeshAccessAddress(randomArray):
        address = Conversion.uint8_array_to_uint32(randomArray)
        bitArray = Conversion.uint32_to_bit_array(address)
        bitArrayAdvertising = Conversion.uint32_to_bit_array(ADVERTISING_ACCESS_ADDRESS)
        
        # Requirement: It shall not be the advertising channel packets’ Access Address.
        if address == ADVERTISING_ACCESS_ADDRESS:
            # print("1")
            return False
        
        # Requirement: It shall not have all four octets equal.
        if randomArray[0] == randomArray[1] and randomArray[0] == randomArray[2] and randomArray[0] == randomArray[3]:
            # print("2")
            return False
    
        # Requirement: It shall not be a sequence that differs from the advertising channel packets’ Access Address by only one bit.
        diffCount = 0
        for idx, bit in enumerate(bitArray):
            if bit != bitArrayAdvertising[idx]:
                diffCount += 1
                
            if diffCount > 1:
                break
        
        if diffCount <= 1:
            # print("3")
            return False
        
        # Requirement: It shall have no more than six consecutive zeros or ones.
        consZero = 0
        consOne = 0
        for bit in bitArray:
            if bit:
                consOne += 1
                consZero = 0
            else:
                consZero += 1
                consOne = 0
                
            if consOne > 5 or consZero > 5:
                break
    
        if consOne > 5 or consZero > 5:
            # print("4")
            return False
    
        # Requirement: It shall have no more than 24 transitions.
        transitions = 0
        for i in range(1, len(bitArray)):
            if bitArray[i] != bitArray[i-1]:
                transitions += 1
        
        if transitions > 24:
            # print("5")
            return False
    
        # Requirement: It shall have a minimum of two transitions in the most significant six bits.
        transitions = 0
        for i in range(1, 6):
            if bitArray[i] != bitArray[i - 1]:
                transitions += 1
        
        if transitions < 2:
            # print("6")
            return False
    
        transitions = 0
        for i in range(len(bitArray)-5, len(bitArray)):
            if bitArray[i] != bitArray[i - 1]:
                transitions += 1
    
        if transitions < 2:
            # print("7")
            return False
        
        return True



from crownstone_core.util.DataStepper import DataStepper
from crownstone_core.protocol.BluenetTypes import PowerSamplesType
from crownstone_core.Exceptions import CrownstoneError, CrownstoneException

class AdcChannelSwapsPacket:
	def __init__(self, data):
		self.count = 0            # uint32
		self.timestamp = 0        # uint32
		self.load(data)

	def load(self, data):
		"""
		Parses data buffer to set member variables.

		data : list of bytes

		Raises exception when parsing fails.
		"""
		streamBuf = DataStepper(data)
		self.count = streamBuf.getUInt32()
		self.timestamp = streamBuf.getUInt32()

	def toString(self):
		msg = "AdcChannelSwapsPacket("
		msg += "count=" + str(self.count)
		msg += " timestamp=" + str(self.timestamp)
		msg += ")"
		return msg

	def __str__(self):
		return self.toString()



class ActiveDays:

    def __init__(self):
        self.Monday = True
        self.Tuesday = True
        self.Wednesday = True
        self.Thursday = True
        self.Friday = True
        self.Saturday = True
        self.Sunday = True

    def fromData(self, data):
        self.Sunday = (data >> 0) & 0x01 == 1
        self.Monday = (data >> 1) & 0x01 == 1
        self.Tuesday = (data >> 2) & 0x01 == 1
        self.Wednesday = (data >> 3) & 0x01 == 1
        self.Thursday = (data >> 4) & 0x01 == 1
        self.Friday = (data >> 5) & 0x01 == 1
        self.Saturday = (data >> 6) & 0x01 == 1
        return self

    def getMask(self):
        mask = 0

        # bits:
        MondayBit = 0
        TuesdayBit = 0
        WednesdayBit = 0
        ThursdayBit = 0
        FridayBit = 0
        SaturdayBit = 0
        SundayBit = 0
        if self.Sunday:  MondayBit = 1
        if self.Monday:  TuesdayBit = 1
        if self.Tuesday:  WednesdayBit = 1
        if self.Wednesday:  ThursdayBit = 1
        if self.Thursday:  FridayBit = 1
        if self.Friday:  SaturdayBit = 1
        if self.Saturday:  SundayBit = 1

        # configure mask
        mask = mask | SundayBit << 0
        mask = mask | MondayBit << 1
        mask = mask | TuesdayBit << 2
        mask = mask | WednesdayBit << 3
        mask = mask | ThursdayBit << 4
        mask = mask | FridayBit << 5
        mask = mask | SaturdayBit << 6

        return mask

    def getDictionary(self):
        returnDict = {
            "Sun": self.Sunday,
            "Mon": self.Monday,
            "Tue": self.Tuesday,
            "Wed": self.Wednesday,
            "Thu": self.Thursday,
            "Fri": self.Friday,
            "Sat": self.Saturday,
        }

        return returnDict



from enum import IntEnum

from crownstone_core.util.Conversion import Conversion
from crownstone_core.util.DataStepper import DataStepper

DEFAULT_PRESENCE_DELAY = 300 # seconds = 5 minutes

class BehaviourPresenceType(IntEnum):
    ignorePresence   = 0
    somoneInLocation = 1
    nobodyInLocation = 2
    someoneInSphere  = 3
    nobodyInSphere   = 4

    @classmethod
    def has_value(cls, value):
        return any(value == item.value for item in cls)


class BehaviourPresence:

    def __init__(self):
        self.presenceType = BehaviourPresenceType.ignorePresence
        self.locationIds = []
        self.delayInSeconds = 0
        self.valid = True

    def setSpherePresence(self, presenceType, delayInSeconds=DEFAULT_PRESENCE_DELAY):
        self.presenceType = presenceType
        self.delayInSeconds = delayInSeconds
        return self

    def setLocationPresence(self, presenceType, locationIds, delayInSeconds=DEFAULT_PRESENCE_DELAY):
        self.presenceType = presenceType
        self.locationIds = locationIds
        self.delayInSeconds = delayInSeconds
        return self

    def fromData(self, data):
        if len(data) != 13:
            self.valid = False
            return self

        payload = DataStepper(data)
        firstByte = payload.getUInt8()
        if not BehaviourPresenceType.has_value(firstByte):
            self.valid = False
            return self

        self.behaviourType = BehaviourPresenceType(firstByte)
        self.locationIds = self.unpackMask(payload.getUInt64())
        self.delayInSeconds = payload.getUInt32()

        return self

    def getMask(self, locationIds):
        result = 0
        bit = 1
        for locationId in locationIds:
            if locationId < 64:
                result = result | bit << locationId

        return result

    def unpackMask(self, mask):
        result = []
        bit = 1
        for i in range(0, 64):
            if (mask >> i & bit) == 1:
                result.append(i)

        return result

    def getPacket(self):
        arr = []

        arr.append(self.presenceType.value)
        arr += Conversion.uint64_to_uint8_array(self.getMask(self.locationIds))
        arr += Conversion.uint32_to_uint8_array(self.delayInSeconds)

        return arr

    def getDictionary(self):
        returnDict = {}

        if self.presenceType == BehaviourPresenceType.ignorePresence:
            returnDict["type"] = "IGNORE"

        elif self.presenceType == BehaviourPresenceType.somoneInLocation:
            returnDict["type"] = "SOMEBODY"
            returnDict["data"] = {"type": "LOCATION", "locationIds": self.locationIds}

        elif self.presenceType == BehaviourPresenceType.someoneInSphere:
            returnDict["type"] = "SOMEBODY"
            returnDict["data"] = {"type": "SPHERE"}

        elif self.presenceType == BehaviourPresenceType.nobodyInLocation:
            returnDict["type"] = "NOBODY"
            returnDict["data"] = {"type": "LOCATION", "locationIds": self.locationIds}

        elif self.presenceType == BehaviourPresenceType.nobodyInSphere:
            returnDict["type"] = "NOBODY"
            returnDict["data"] = {"type": "SPHERE"}

        if self.presenceType != BehaviourPresenceType.ignorePresence:
            returnDict["delay"] = self.delayInSeconds

        return returnDict




from enum import Enum


class EncryptionError(Enum):
    CAN_NOT_FIND_SERVICE              = "CAN_NOT_FIND_SERVICE"
    CAN_NOT_FIND_CHACTERISTIC         = "CAN_NOT_FIND_CHACTERISTIC"
    CAN_NOT_GET_CHARACTERISTIC        = "CAN_NOT_GET_CHARACTERISTIC"
    CAN_NOT_FIND_CCCD                 = "CAN_NOT_FIND_CCCD"
    ABORT_NOTIFICATION_STREAM_W_ERROR = "ABORT_NOTIFICATION_STREAM_W_ERROR"
    NOTIFICATION_STREAM_TIMEOUT       = "NOTIFICATION_STREAM_TIMEOUT"
    NO_NOTIFICATION_DATA_RECEIVED     = "NO_NOTIFICATION_DATA_RECEIVED"
    INVALID_SESSION_NONCE             = "INVALID_SESSION_NONCE"
    INVALID_SESSION_DATA              = "INVALID_SESSION_DATA"
    INVALID_ENCRYPTION_PACKAGE        = "INVALID_ENCRYPTION_PACKAGE"
    INVALID_ENCRYPTION_USER_LEVEL     = "INVALID_ENCRYPTION_USER_LEVEL"
    COULD_NOT_VALIDATE_SESSION_NONCE  = "COULD_NOT_VALIDATE_SESSION_NONCE"
    COULD_NOT_READ_SESSION_NONCE      = "COULD_NOT_READ_SESSION_NONCE"
    NO_SESSION_NONCE_SET              = "NO_SESSION_NONCE_SET"
    NO_ENCRYPTION_KEYS_SET            = "NO_ENCRYPTION_KEYS_SET"
    ENCRYPTION_VALIDATION_FAILED      = "ENCRYPTION_VALIDATION_FAILED"
    SETUP_FAILED                      = "SETUP_FAILED"
    NOT_IN_RECOVERY_MODE              = "NOT_IN_RECOVERY_MODE"
    RECOVERY_MODE_DISABLED            = "RECOVERY_MODE_DISABLED"


class CrownstoneError(Enum):
    ADMIN_KEY_REQUIRED                = "ADMIN_KEY_REQUIRED"
    MEMBER_KEY_REQUIRED               = "MEMBER_KEY_REQUIRED"
    BASIC_KEY_REQUIRED                = "BASIC_KEY_REQUIRED"
    SERVICE_DATA_KEY_REQUIRED         = "SERVICE_DATA_KEY_REQUIRED"
    LOCALIZATION_KEY_REQUIRED         = "LOCALIZATION_KEY_REQUIRED"
    MESH_APP_KEY                      = "MESH_APP_KEY"
    MESH_NETWORK_KEY                  = "MESH_NETWORK_KEY"
    INVALID_ADDRESS                   = "INVALID_ADDRESS"

    NOT_IMPLEMENTED_YET               = "NOT_IMPLEMENTED_YET"

    DISCONNECTED                                        = "DISCONNECTED"
    CONNECTION_CANCELLED                                = "CONNECTION_CANCELLED"
    CONNECTION_FAILED                                   = "CONNECTION_FAILED"
    NOT_CONNECTED                                       = "NOT_CONNECTED"
    NO_SERVICES                                         = "NO_SERVICES"
    NO_CHARACTERISTICS                                  = "NO_CHARACTERISTICS"
    SERVICE_DOES_NOT_EXIST                              = "SERVICE_DOES_NOT_EXIST"
    CHARACTERISTIC_DOES_NOT_EXIST                       = "CHARACTERISTIC_DOES_NOT_EXIST"
    WRONG_TYPE_OF_PROMISE                               = "WRONG_TYPE_OF_PROMISE"
    INVALID_UUID                                        = "INVALID_UUID"
    NOT_INITIALIZED                                     = "NOT_INITIALIZED"
    CANNOT_SET_TIMEOUT_WITH_THIS_TYPE_OF_PROMISE        = "CANNOT_SET_TIMEOUT_WITH_THIS_TYPE_OF_PROMISE"
    TIMEOUT                                             = "TIMEOUT"
    DISCONNECT_TIMEOUT                                  = "DISCONNECT_TIMEOUT"
    ERROR_DISCONNECT_TIMEOUT                            = "ERROR_DISCONNECT_TIMEOUT"
    AWAIT_DISCONNECT_TIMEOUT                            = "AWAIT_DISCONNECT_TIMEOUT"
    CANCEL_PENDING_CONNECTION_TIMEOUT                   = "CANCEL_PENDING_CONNECTION_TIMEOUT"
    CONNECT_TIMEOUT                                     = "CONNECT_TIMEOUT"
    GET_SERVICES_TIMEOUT                                = "GET_SERVICES_TIMEOUT"
    GET_CHARACTERISTICS_TIMEOUT                         = "GET_CHARACTERISTICS_TIMEOUT"
    READ_CHARACTERISTIC_TIMEOUT                         = "READ_CHARACTERISTIC_TIMEOUT"
    WRITE_CHARACTERISTIC_TIMEOUT                        = "WRITE_CHARACTERISTIC_TIMEOUT"
    ENABLE_NOTIFICATIONS_TIMEOUT                        = "ENABLE_NOTIFICATIONS_TIMEOUT"
    NOTIFICATION_STREAM_TIMEOUT                         = "NOTIFICATION_STREAM_TIMEOUT"
    DISABLE_NOTIFICATIONS_TIMEOUT                       = "DISABLE_NOTIFICATIONS_TIMEOUT"
    CANNOT_WRITE_AND_VERIFY                             = "CANNOT_WRITE_AND_VERIFY"
    CAN_NOT_CONNECT_TO_UUID                             = "CAN_NOT_CONNECT_TO_UUID"
    COULD_NOT_FACTORY_RESET                             = "COULD_NOT_FACTORY_RESET"
    INCORRECT_RESPONSE_LENGTH                           = "INCORRECT_RESPONSE_LENGTH"
    UNKNOWN_TYPE                                        = "UNKNOWN_TYPE"
    COULD_NOT_GET_LOCATION                              = "COULD_NOT_GET_LOCATION"
    INVALID_SESSION_REFERENCE_ID                        = "INVALID_SESSION_REFERENCE_ID"
    INVALID_SESSION_DATA                                = "INVALID_SESSION_DATA"
    NO_SESSION_NONCE_SET                                = "NO_SESSION_NONCE_SET"
    COULD_NOT_VALIDATE_SESSION_NONCE                    = "COULD_NOT_VALIDATE_SESSION_NONCE"
    INVALID_SIZE_FOR_ENCRYPTED_PAYLOAD                  = "INVALID_SIZE_FOR_ENCRYPTED_PAYLOAD"
    INVALID_SIZE_FOR_SESSION_NONCE_PACKET               = "INVALID_SIZE_FOR_SESSION_NONCE_PACKET"
    INVALID_PACKAGE_FOR_ENCRYPTION_TOO_SHORT            = "INVALID_PACKAGE_FOR_ENCRYPTION_TOO_SHORT"
    INVALID_KEY_FOR_ENCRYPTION                          = "INVALID_KEY_FOR_ENCRYPTION"
    DO_NOT_HAVE_ENCRYPTION_KEY                          = "DO_NOT_HAVE_ENCRYPTION_KEY"
    COULD_NOT_ENCRYPT                                   = "COULD_NOT_ENCRYPT"
    COULD_NOT_ENCRYPT_KEYS_NOT_SET                      = "COULD_NOT_ENCRYPT_KEYS_NOT_SET"
    COULD_NOT_DECRYPT_KEYS_NOT_SET                      = "COULD_NOT_DECRYPT_KEYS_NOT_SET"
    COULD_NOT_DECRYPT                                   = "COULD_NOT_DECRYPT"
    CAN_NOT_GET_PAYLOAD                                 = "CAN_NOT_GET_PAYLOAD"
    USERLEVEL_IN_READ_PACKET_INVALID                    = "USERLEVEL_IN_READ_PACKET_INVALID"
    READ_SESSION_NONCE_ZERO_MAYBE_ENCRYPTION_DISABLED   = "READ_SESSION_NONCE_ZERO_MAYBE_ENCRYPTION_DISABLED"
    SETUP_FAILED                                        = "SETUP_FAILED"
    NOT_IN_RECOVERY_MODE                                = "NOT_IN_RECOVERY_MODE"
    CANNOT_READ_FACTORY_RESET_CHARACTERISTIC            = "CANNOT_READ_FACTORY_RESET_CHARACTERISTIC"
    RECOVER_MODE_DISABLED                               = "RECOVER_MODE_DISABLED"
    INVALID_TX_POWER_VALUE                              = "INVALID_TX_POWER_VALUE"
    NO_KEEPALIVE_STATE_ITEMS                            = "NO_KEEPALIVE_STATE_ITEMS"
    NO_SWITCH_STATE_ITEMS                               = "NO_SWITCH_STATE_ITEMS"
    DFU_OVERRULED                                       = "DFU_OVERRULED"
    DFU_ABORTED                                         = "DFU_ABORTED"
    DFU_ERROR                                           = "DFU_ERROR"
    COULD_NOT_FIND_PERIPHERAL                           = "COULD_NOT_FIND_PERIPHERAL"
    PACKETS_DO_NOT_MATCH                                = "PACKETS_DO_NOT_MATCH"
    NOT_IN_DFU_MODE                                     = "NOT_IN_DFU_MODE"
    REPLACED_WITH_OTHER_PROMISE                         = "REPLACED_WITH_OTHER_PROMISE"
    BLE_RESET                                           = "BLE_RESET"
    INCORRECT_SCHEDULE_ENTRY_INDEX                      = "INCORRECT_SCHEDULE_ENTRY_INDEX"
    INCORRECT_DATA_COUNT_FOR_ALL_TIMERS                 = "INCORRECT_DATA_COUNT_FOR_ALL_TIMERS"
    NO_SCHEDULE_ENTRIES_AVAILABLE                       = "NO_SCHEDULE_ENTRIES_AVAILABLE"
    NO_TIMER_FOUND                                      = "NO_TIMER_FOUND"
    PROCESS_ABORTED_WITH_ERROR                          = "PROCESS_ABORTED_WITH_ERROR"
    UNKNOWN_PROCESS_TYPE                                = "UNKNOWN_PROCESS_TYPE"
    INVALID_INPUT                                       = "INVALID_INPUT"
    INVALID_BROADCAST_ACCESS_LEVEL                      = "INVALID_BROADCAST_ACCESS_LEVEL"
    INVALID_BROADCAST_LOCATION_ID                       = "INVALID_BROADCAST_LOCATION_ID"
    INVALID_BROADCAST_PROFILE_INDEX                     = "INVALID_BROADCAST_PROFILE_INDEX"
    INVALID_BROADCAST_PAYLOAD_SIZE                      = "INVALID_BROADCAST_PAYLOAD_SIZE"
    BROADCAST_ERROR                                     = "BROADCAST_ERROR"
    BROADCAST_ABORTED                                   = "BROADCAST_ABORTED"
    BEHAVIOUR_INDEX_OUT_OF_RANGE                        = "BEHAVIOUR_INDEX_OUT_OF_RANGE"
    BEHAVIOUR_INVALID                                   = "BEHAVIOUR_INVALID"
    BEHAVIOUR_INVALID_RESPONSE                          = "BEHAVIOUR_INVALID_RESPONSE"
    BEHAVIOUR_NOT_FOUND_AT_INDEX                        = "BEHAVIOUR_NOT_FOUND_AT_INDEX"
    PROFILE_INDEX_MISSING                               = "PROFILE_INDEX_MISSING"
    TYPE_MISSING                                        = "TYPE_MISSING"
    DATA_MISSING                                        = "DATA_MISSING"
    ACTIVE_DAYS_MISSING                                 = "ACTIVE_DAYS_MISSING"
    ACTIVE_DAYS_INVALID                                 = "ACTIVE_DAYS_INVALID"
    NO_ACTIVE_DAYS                                      = "NO_ACTIVE_DAYS"
    BEHAVIOUR_ACTION_MISSING                            = "BEHAVIOUR_ACTION_MISSING"
    BEHAVIOUR_TIME_MISSING                              = "BEHAVIOUR_TIME_MISSING"
    BEHAVIOUR_INTENSITY_MISSING                         = "BEHAVIOUR_INTENSITY_MISSING"
    TWILIGHT_CANT_HAVE_PRESENCE                         = "TWILIGHT_CANT_HAVE_PRESENCE"
    TWILIGHT_CANT_HAVE_END_CONDITION                    = "TWILIGHT_CANT_HAVE_END_CONDITION"
    NO_TIME_TYPE                                        = "NO_TIME_TYPE"
    INVALID_TIME_TYPE                                   = "INVALID_TIME_TYPE"
    MISSING_TO_TIME                                     = "MISSING_TO_TIME"
    MISSING_FROM_TIME                                   = "MISSING_FROM_TIME"
    MISSING_TO_TIME_TYPE                                = "MISSING_TO_TIME_TYPE"
    MISSING_FROM_TIME_DATA                              = "MISSING_FROM_TIME_DATA"
    MISSING_TO_TIME_DATA                                = "MISSING_TO_TIME_DATA"
    MISSING_FROM_TIME_TYPE                              = "MISSING_FROM_TIME_TYPE"
    INVALID_TIME_FROM_TYPE                              = "INVALID_TIME_FROM_TYPE"
    INVALID_TIME_TO_TYPE                                = "INVALID_TIME_TO_TYPE"
    INVALID_FROM_DATA                                   = "INVALID_FROM_DATA"
    INVALID_TO_DATA                                     = "INVALID_TO_DATA"
    INVALID_PRESENCE_TYPE                               = "INVALID_PRESENCE_TYPE"
    NO_PRESENCE_TYPE                                    = "NO_PRESENCE_TYPE"
    NO_PRESENCE_DATA                                    = "NO_PRESENCE_DATA"
    NO_PRESENCE_DELAY                                   = "NO_PRESENCE_DELAY"
    NO_PRESENCE_LOCATION_IDS                            = "NO_PRESENCE_LOCATION_IDS"
    NO_END_CONDITION_TYPE                               = "NO_END_CONDITION_TYPE"
    NO_END_CONDITION_PRESENCE                           = "NO_END_CONDITION_PRESENCE"
    NO_END_CONDITION_DURATION                           = "NO_END_CONDITION_DURATION"
    PROTOCOL_NOT_SUPPORTED                              = "PROTOCOL_NOT_SUPPORTED"
    RESULT_NOT_SUCCESS                                  = "RESULT_NOT_SUCCESS"


class CrownstoneBleException(Exception):
    code    = None
    type    = None
    message = None
    
    def __init__(self, type, message="", code=0):
        self.type = type
        self.message = message
        self.code = code
        

class CrownstoneException(Exception):
    code    = None
    type    = None
    message = None
    
    def __init__(self, type, message, code=0):
        self.type = type
        self.message = message
        self.code = code



import random

import math
import pyaes

from crownstone_core.Constants import UserLevel
from crownstone_core.Exceptions import CrownstoneBleException, EncryptionError
from crownstone_core.util.Conversion import Conversion

BLOCK_LENGTH             = 16
NONCE_LENGTH             = 16
SESSION_DATA_LENGTH      = 5
SESSION_KEY_LENGTH       = 4
PACKET_USER_LEVEL_LENGTH = 1
PACKET_NONCE_LENGTH      = 3
CHECKSUM                 = 0xcafebabe




class EncryptedPackage:
    
    def __init__(self, dataArray):
        self.nonce = None
        self.userLevel = None
        self.payload = None
        
        prefixLength = PACKET_NONCE_LENGTH + PACKET_USER_LEVEL_LENGTH
        # 20 is the minimal size of a packet (3+1+16)
        if len(dataArray) < 20:
            raise CrownstoneBleException(EncryptionError.INVALID_ENCRYPTION_PACKAGE, "Invalid package for encryption. It is too short (min length 20) got " + str(len(dataArray)) + " bytes.")

        self.nonce = [0] * PACKET_NONCE_LENGTH
        
        for i in range(0, PACKET_NONCE_LENGTH):
            self.nonce[i] = dataArray[i]
            
        if dataArray[PACKET_NONCE_LENGTH] > 2 and dataArray[PACKET_NONCE_LENGTH] != UserLevel.setup.value:
            raise CrownstoneBleException(EncryptionError.INVALID_ENCRYPTION_USER_LEVEL, "User level in read packet is invalid:" + str(dataArray[PACKET_NONCE_LENGTH]))
        
        try:
            self.userLevel = UserLevel(dataArray[PACKET_NONCE_LENGTH])
        except ValueError:
            raise CrownstoneBleException(EncryptionError.INVALID_ENCRYPTION_USER_LEVEL, "User level in read packet is invalid:" + str(dataArray[PACKET_NONCE_LENGTH]))
        
        payload = [0] * (len(dataArray) - prefixLength)
        for i in range(0, (len(dataArray) - prefixLength)):
            payload[i] = dataArray[i + prefixLength]
            
        if len(payload) % 16 != 0:
            raise CrownstoneBleException(EncryptionError.INVALID_ENCRYPTION_PACKAGE, "Invalid size for encrypted payload")
        
        self.payload = payload


class EncryptionHandler:
    

    @staticmethod
    def decryptECB(uint8Array, key):
        aes = pyaes.AESModeOfOperationECB(key)
        
        stringPayload = "".join(chr(b) for b in uint8Array)
        
        decrypted = aes.decrypt(stringPayload)
        
        return decrypted


    @staticmethod
    def encryptECB(uint8Array, key):
        aes = pyaes.AESModeOfOperationECB(key)
    
        stringPayload = "".join(chr(b) for b in uint8Array)
    
        encrypted = aes.encrypt(stringPayload)
    
        return encrypted


    @staticmethod
    def decryptCTR(data, packetNonce, sessionNonce, key):
        IV = EncryptionHandler.generateIV(packetNonce, sessionNonce)
    
        stringPayload = "".join(chr(b) for b in data)
    
        aes = pyaes.AESModeOfOperationCTR(key, counter=IVCounter(IV))
    
        decryptedData = aes.decrypt(stringPayload)
    
        return decryptedData


    @staticmethod
    def decrypt(data, settings):
        if settings.sessionNonce is None:
            raise CrownstoneBleException(EncryptionError.NO_SESSION_NONCE_SET, "Can't Decrypt: No session nonce set")
    
        if settings.userLevel == UserLevel.unknown:
            raise CrownstoneBleException(EncryptionError.NO_ENCRYPTION_KEYS_SET, "Can't Decrypt: No encryption keys set.")
        
        #unpack the session data
        package = EncryptedPackage(data)
        
        key = EncryptionHandler._getKeyForLevel(package.userLevel, settings)
    
        # decrypt data
        decrypted = EncryptionHandler.decryptCTR(package.payload, package.nonce, settings.sessionNonce, key)
        
        return EncryptionHandler._verifyDecryption(decrypted, settings.validationKey)
    
    
    @staticmethod
    def _verifyDecryption(decrypted, validationKey):
        # the conversion to uint32 only takes the first 4 bytes
        if Conversion.uint8_array_to_uint32(decrypted) == Conversion.uint8_array_to_uint32(validationKey):
            # remove checksum from decryption and return payload
            result = [0] * (len(decrypted) - SESSION_KEY_LENGTH)
            for i in range(0,len(result)):
                result[i] = decrypted[i+SESSION_KEY_LENGTH]
            return result
    
        else:
            raise CrownstoneBleException(EncryptionError.ENCRYPTION_VALIDATION_FAILED, "Failed to validate result, Could not decrypt")
            
    
    @staticmethod
    def getRandomNumber(testing=False):
        if testing:
            return 128
        return random.randint(0,255)


    @staticmethod
    def encryptCTR(dataArray, packetNonce, settings, key):

        IV = EncryptionHandler.generateIV(packetNonce, settings.sessionNonce)
        
        # calculate the amount of blocks
        amountOfBlocks = int(math.ceil(float(SESSION_KEY_LENGTH + len(dataArray)) / float(BLOCK_LENGTH)))
    
        # create buffer that is zero padded
        paddedPayload = [0] * amountOfBlocks * BLOCK_LENGTH
    
        # fill the payload with the key and the data
        for i in range(0, SESSION_KEY_LENGTH):
            paddedPayload[i] = settings.validationKey[i]
    
        for i in range(0, len(dataArray)):
            paddedPayload[i + SESSION_KEY_LENGTH] = dataArray[i]
    
        stringPayload = "".join(chr(b) for b in paddedPayload)
    
        aes = pyaes.AESModeOfOperationCTR(key, counter=IVCounter(IV))
    
        encryptedData = aes.encrypt(stringPayload)
        
        return encryptedData
    
    @staticmethod
    def encrypt(dataArray, settings):
        if settings.sessionNonce is None:
            raise CrownstoneBleException(EncryptionError.NO_SESSION_NONCE_SET, "Can't Decrypt: No session nonce set")
    
        if settings.userLevel == UserLevel.unknown:
            raise CrownstoneBleException(EncryptionError.NO_ENCRYPTION_KEYS_SET, "Can't Decrypt: No encryption keys set.")

        packetNonce = [0] * PACKET_NONCE_LENGTH
        # create a random nonce
        for i in range(0, PACKET_NONCE_LENGTH):
            packetNonce[i] = EncryptionHandler.getRandomNumber()
        
        key = EncryptionHandler._getKey(settings)
        encryptedData = EncryptionHandler.encryptCTR(dataArray, packetNonce, settings, key)
    
        result = packetNonce + [settings.userLevel.value]
        
        for byte in encryptedData:
            result.append(byte)
    
        return bytes(result)
            
    
    @staticmethod
    def _getKey(settings):
        return EncryptionHandler._getKeyForLevel(settings.userLevel, settings)
        
    @staticmethod
    def _getKeyForLevel(userLevel, settings):
        if settings.initializedKeys == False and userLevel != UserLevel.setup:
            raise CrownstoneBleException(EncryptionError.NO_ENCRYPTION_KEYS_SET, "Could not encrypt: Keys not set.")
    
        key = None
        if userLevel == UserLevel.admin:
            key = settings.adminKey
        elif userLevel == UserLevel.member:
            key = settings.memberKey
        elif userLevel == UserLevel.basic:
            key = settings.basicKey
        elif userLevel == UserLevel.setup:
            key = settings.setupKey
        else:
            raise CrownstoneBleException(EncryptionError.NO_ENCRYPTION_KEYS_SET, "Could not encrypt: Invalid key for encryption.")
    
        if key is None:
            raise CrownstoneBleException(EncryptionError.NO_ENCRYPTION_KEYS_SET, "Could not encrypt: Keys not set.")
    
        return key
        
    
    @staticmethod
    def generateIV(packetNonce, sessionData):
        if len(packetNonce) != PACKET_NONCE_LENGTH:
            raise CrownstoneBleException(EncryptionError.INVALID_SESSION_NONCE, "Invalid size for session nonce packet")
        
        IV = [0] * NONCE_LENGTH
        
        # the IV used in the CTR mode is 8 bytes, the first 3 are random
        for i in range(0,PACKET_NONCE_LENGTH):
            IV[i] = packetNonce[i]
        
        # the IV used in the CTR mode is 8 bytes, the last 5 are from the session data
        for i in range(0,SESSION_DATA_LENGTH):
            IV[i + PACKET_NONCE_LENGTH] = sessionData[i]
            
        return IV

    """
     * This method is used to encrypt data with the CTR method and wrap the envelope around it according to protocol V5
    """
    @staticmethod
    def encryptBroadcast(dataArray, key, IV, validationNonce):
        IV = IV + [0] * (NONCE_LENGTH - len(IV))

        # create buffer that is zero padded
        paddedPayload = [0] * BLOCK_LENGTH

        # fill the payload with the key and the data
        for i in range(0, SESSION_KEY_LENGTH):
            paddedPayload[i] = validationNonce[i]

        for i in range(0, len(dataArray)):
            paddedPayload[i + SESSION_KEY_LENGTH] = dataArray[i]

        stringPayload = "".join(chr(b) for b in paddedPayload)

        aes = pyaes.AESModeOfOperationCTR(key, counter=IVCounter(IV))

        encryptedData = aes.encrypt(stringPayload)

        return encryptedData


    
class IVCounter(object):
    """
        A counter object for the Counter (CTR) mode of operation.

       To create a custom counter, you can usually just override the
       increment method.
    """
    
    def __init__(self, initialList):
        
        # Convert the value into an array of bytes long
        self._counter = initialList

    def get_value(self):
        return self._counter

    value = property(get_value)
    
    def increment(self):
        self._counter[len(self._counter)-1] += 1

        


"""
Copied from https://raw.githubusercontent.com/buckley-w-david/python-crypto/master/python_crypto/symmetric/rc5.py
MIT Licensed

Minor modifications
"""


import typing

from crownstone_core.util.Conversion import Conversion


def _rotate_left(val: int, r_bits: int, max_bits: int) -> int:
    v1 = (val << r_bits % max_bits) & (2 ** max_bits - 1)
    v2 = (val & (2 ** max_bits - 1)) >> (max_bits - (r_bits % max_bits))
    return v1 | v2


def _rotate_right(val: int, r_bits: int, max_bits: int) -> int:
    v1 = (val & (2 ** max_bits - 1)) >> r_bits % max_bits
    v2 = val << (max_bits - (r_bits % max_bits)) & (2 ** max_bits - 1)

    return v1 | v2


def _expand_key(key: bytes, wordsize: int, rounds: int) -> typing.List[int]:
    # Pads key so that it is aligned with the word size, then splits it into words
    def _align_key(key: bytes, align_val: int) -> typing.List[int]:
        while len(key) % align_val:
            key += (
                b"\x00"
            )  # Add 0 bytes until the key length is aligned to the block size

        L = []
        for i in range(0, len(key), align_val):
            L.append(int.from_bytes(key[i : i + align_val], byteorder="little"))

        return L

    # generation function of the constants for the extend step
    def _const(w: int) -> typing.Tuple[int, int]:
        if w == 16:
            return 0xB7E1, 0x9E37  # Returns the value of P and Q
        elif w == 32:
            return 0xB7E15163, 0x9E3779B9
        elif w == 64:
            return 0xB7E151628AED2A6B, 0x9E3779B97F4A7C15
        raise ValueError("Bad word sie")

    # Generate pseudo-random list S
    def _extend_key(w: int, r: int) -> typing.List[int]:
        P, Q = _const(w)
        S = [P]
        t = 2 * (r + 1)
        for i in range(1, t):
            S.append((S[i - 1] + Q) % 2 ** w)

        return S

    def _mix(L: typing.List[int], S: typing.List[int], r: int, w: int, c: int) -> typing.List[int]:
        t = 2 * (r + 1)
        m = max(c, t)
        A = B = i = j = 0

        for k in range(3 * m):
            A = S[i] = _rotate_left(S[i] + A + B, 3, w)
            B = L[j] = _rotate_left(L[j] + A + B, A + B, w)

            i = (i + 1) % t
            j = (j + 1) % c

        return S

    aligned = _align_key(key, wordsize // 8)
    extended = _extend_key(wordsize, rounds)

    S = _mix(aligned, extended, rounds, wordsize, len(aligned))

    return S


def _encrypt_block(data: bytes, expanded_key: typing.List[int], blocksize: int, rounds: int) -> bytes:
    w = blocksize // 2
    b = blocksize // 8
    mod = 2 ** w

    A = int.from_bytes(data[: b // 2], byteorder="little")
    B = int.from_bytes(data[b // 2 :], byteorder="little")

    A = (A + expanded_key[0]) % mod
    B = (B + expanded_key[1]) % mod

    for i in range(1, rounds + 1):
        A = (_rotate_left((A ^ B), B, w) + expanded_key[2 * i]) % mod
        B = (_rotate_left((A ^ B), A, w) + expanded_key[2 * i + 1]) % mod

    res = A.to_bytes(b // 2, byteorder="little") + B.to_bytes(b // 2, byteorder="little")
    return res


def _decrypt_block(data: bytes, expanded_key: typing.List[int], blocksize: int, rounds: int) -> bytes:
    w = blocksize // 2
    b = blocksize // 8
    mod = 2 ** w

    A = int.from_bytes(data[: b // 2], byteorder="little")
    B = int.from_bytes(data[b // 2 :], byteorder="little")

    for i in range(rounds, 0, -1):
        B = _rotate_right(B - expanded_key[2 * i + 1], A, w) ^ A
        A = _rotate_right((A - expanded_key[2 * i]), B, w) ^ B

    B = (B - expanded_key[1]) % mod
    A = (A - expanded_key[0]) % mod

    res = A.to_bytes(b // 2, byteorder="little") + B.to_bytes(b // 2, byteorder="little")
    return res


class RC5:
    def __init__(self, key: bytes, blocksize: int, rounds: int) -> None:
        self.key = key
        self.blocksize = blocksize
        self.rounds = rounds

    def encrypt(self, data: int) -> int:
        # alteration to accept numbers instead of bytes
        data = Conversion.uint32_to_uint16_reversed_array(data)
        data = bytes(Conversion.uint16_array_to_uint8_array(data))

        blocksize = self.blocksize
        key = self.key
        rounds = self.rounds

        w = blocksize // 2
        b = blocksize // 8

        expanded_key = _expand_key(key, w, rounds)

        index = b
        chunk = data[:index]
        out = []
        while chunk:
            chunk = chunk.ljust(b, b"\x00")  # padding with 0 bytes if not large enough
            encrypted_chunk = _encrypt_block(chunk, expanded_key, blocksize, rounds)
            out.append(encrypted_chunk)

            chunk = data[index : index + b]  # Read in blocksize number of bytes
            index += b
        result = b"".join(out)

        result = Conversion.uint8_array_to_uint16_array(result)
        result = Conversion.uint16_reversed_array_to_uint32(result)

        return result

    def decrypt(self, data: int) -> bytes:
        data = Conversion.uint32_to_uint16_reversed_array(data)
        data = bytes(Conversion.uint16_array_to_uint8_array(data))

        blocksize = self.blocksize
        key = self.key
        rounds = self.rounds

        w = blocksize // 2
        b = blocksize // 8

        expanded_key = _expand_key(key, w, rounds)

        index = b
        chunk = data[:index]
        out = []
        while chunk:
            decrypted_chunk = _decrypt_block(chunk, expanded_key, blocksize, rounds)
            chunk = data[index : index + b]  # Read in blocksize number of bytes
            if not chunk:
                decrypted_chunk = decrypted_chunk.rstrip(b"\x00")

            index += b
            out.append(decrypted_chunk)
        result = b"".join(out)

        result = Conversion.uint8_array_to_uint16_array(result)
        result = Conversion.uint16_reversed_array_to_uint32(result)

        return result


from enum import IntEnum


class ControlType(IntEnum):
    SETUP                      = 0
    FACTORY_RESET              = 1
    GET_STATE                  = 2
    SET_STATE                  = 3
    GET_BOOTLOADER_VERSION     = 4
    GET_UICR_DATA              = 5
    SET_IBEACON_CONFIG_ID      = 6
    RESET                      = 10
    GOTO_DFU                   = 11
    NO_OPERATION               = 12
    DISCONNECT                 = 13
    SWITCH                     = 20
    MULTISWITCH                = 21
    PWM                        = 22
    RELAY                      = 23
    SET_TIME                   = 30
    SET_TX                     = 31
    RESET_ERRORS               = 32
    MESH_COMMAND               = 33
    SET_SUN_TIME               = 34
    ALLOW_DIMMING              = 40
    LOCK_SWITCH                = 41
    UART_MESSAGE               = 50
    SAVE_BEHAVIOUR             = 60
    REPLACE_BEHAVIOUR          = 61
    REMOVE_BEHAVIOUR           = 62
    GET_BEHAVIOUR              = 63
    GET_BEHAVIOUR_INDICES      = 64
    BEHAVIOUR_HANDLER_SETTINGS = 65
    GET_BEHAVIOUR_DEBUG        = 69
    REGISTER_TRACKED_DEVICE    = 70
    TRACKED_DEVICE_HEARTBEAT   = 71
    GET_UPTIME                 = 80
    GET_ADC_RESTARTS           = 81
    GET_SWITCH_HISTORY         = 82
    GET_POWER_SAMPLES          = 83
    GET_MIN_SCHEDULER_FREE     = 84
    GET_LAST_RESET_REASON      = 85
    GET_GPREGRET               = 86
    GET_ADC_CHANNEL_SWAPS      = 87
    GET_RAM_STATS              = 88
    MICROAPP                   = 90
    UNSPECIFIED                = 65535

    @classmethod
    def has_value(cls, value):
        return any(value == item.value for item in cls)


class StateType(IntEnum):
    PWM_PERIOD                             = 5
    IBEACON_MAJOR                          = 6
    IBEACON_MINOR                          = 7
    IBEACON_UUID                           = 8
    IBEACON_TX_POWER                       = 9
    WIFI_SETTINGS                          = 10
    TX_POWER                               = 11
    ADVERTISEMENT_INTERVAL                 = 12
    PASSKEY                                = 13
    MIN_ENV_TEMP                           = 14
    MAX_ENV_TEMP                           = 15
    SCAN_DURATION                          = 16
    SCAN_SEND_DELAY                        = 17
    SCAN_BREAK_DURATION                    = 18
    BOOT_DELAY                             = 19
    MAX_CHIP_TEMP                          = 20
    SCAN_FILTER                            = 21
    SCAN_FILTER_FRACTION                   = 22
    CURRENT_LIMIT                          = 23
    MESH_ENABLED                           = 24
    ENCRYPTION_ENABLED                     = 25
    IBEACON_ENABLED                        = 26
    SCANNER_ENABLED                        = 27
    CONTINUOUS_POWER_MEASUREMENT_ENABLED   = 28
    TRACKER_ENABLED                        = 29
    ADC_SAMPLE_RATE                        = 30
    POWER_SAMPLE_BURST_INTERVAL            = 31
    POWER_SAMPLE_CONTINUOUS_INTERVAL       = 32
    POWER_SAMPLE_CONTINUOUS_NUMBER_SAMPLES = 33
    CROWNSTONE_IDENTIFIER                  = 34
    ADMIN_ENCRYPTION_KEY                   = 35
    MEMBER_ENCRYPTION_KEY                  = 36
    GUEST_ENCRYPTION_KEY                   = 37
    DEFAULT_ON                             = 38
    SCAN_INTERVAL                          = 39
    SCAN_WINDOW                            = 40
    RELAY_HIGH_DURATION                    = 41
    LOW_TX_POWER                           = 42
    VOLTAGE_MULTIPLIER                     = 43
    CURRENT_MULITPLIER                     = 44
    VOLTAGE_ZERO                           = 45
    CURRENT_ZERO                           = 46
    POWER_ZERO                             = 47
    POWER_AVERAGE_WINDOW                   = 48
    MESH_ACCESS_ADDRESS                    = 49
    CURRENT_CONSUMPTION_THRESHOLD          = 50
    CURRENT_CONSUMPTION_THRESHOLD_DIMMER   = 51
    DIMMER_TEMP_UP_VOLTAGE                 = 52
    DIMMER_TEMP_DOWN_VOLTAGE               = 53
    PWM_ALLOWED                            = 54
    SWITCH_LOCKED                          = 55
    SWITCHCRAFT_ENABLED                    = 56
    SWITCHCRAFT_THRESHOLD                  = 57
    MESH_CHANNEL                           = 58
    UART_ENABLED                           = 59
    DEVICE_NAME                            = 60
    SERVICE_DATA_KEY                       = 61
    MESH_DEVICE_KEY                        = 62
    MESH_APPLICATION_KEY                   = 63
    MESH_NETWORK_KEY                       = 64
    LOCALIZATION_KEY                       = 65
    START_DIMMER_ON_ZERO_CROSSING          = 66
    TAP_TO_TOGGLE_ENABLED                  = 67
    TAP_TO_TOGGLE_RSSI_THRESHOLD_OFFSET    = 68
    BEHAVIOUR_RULE                         = 69
    TWILIGHT_RULE                          = 70
    EXTENDED_BEHAVIOUR_RULE                = 71

    RESET_COUNTER                          = 128
    SWITCH_STATE                           = 129
    ACCUMULATED_ENERGY                     = 130
    POWER_USAGE                            = 131
    TRACKED_DEVICES                        = 132
    SCHEDULE                               = 133
    OPERATION_MODE                         = 134
    TEMPERATURE                            = 135
    TIME                                   = 136
    ERROR_BITMASK                          = 139
    SUNTIMES                               = 149
    BEHAVIOUR_SETTINGS                     = 150

    MESH_IV_INDEX                          = 151
    MESH_SEQ_NUMBER                        = 152
    SOFT_ON_SPEED                          = 156
    HUB_MODE                               = 157
    UART_KEY                               = 158

    @classmethod
    def has_value(cls, value):
        return any(value == item.value for item in cls)

class OpCode(IntEnum):
    READ                       = 0
    WRITE                      = 1
    NOTIFY                     = 2


class DeviceType(IntEnum):
    UNDEFINED                  = 0
    PLUG                       = 1
    GUIDESTONE                 = 2
    BUILTIN                    = 3
    CROWNSTONE_USB             = 4
    BUILTIN_ONE                = 5

    @classmethod
    def has_value(cls, value):
        return any(value == item.value for item in cls)


class ResultValue(IntEnum):
    SUCCESS                    = 0      # Completed successfully.
    WAIT_FOR_SUCCESS           = 1      # Command is successful so far, but you need to wait for SUCCESS.
    BUFFER_UNASSIGNED          = 16     # No buffer was assigned for the command.
    BUFFER_LOCKED              = 17     # Buffer is locked, failed queue command.
    BUFFER_TO_SMALL            = 18
    WRONG_PAYLOAD_LENGTH       = 32     # Wrong payload length provided.
    WRONG_PARAMETER            = 33     # Wrong parameter provided.
    INVALID_MESSAGE            = 34     # invalid message provided.
    UNKNOWN_OP_CODE            = 35     # Unknown operation code provided.
    UNKNOWN_TYPE               = 36     # Unknown type provided.
    NOT_FOUND                  = 37     # The thing you were looking for was not found.
    NO_SPACE                   = 38
    BUSY                       = 39
    ERR_WRONG_STATE            = 40     # The crownstone is in a wrong state.
    ERR_ALREADY_EXISTS         = 41     # Item already exists.
    ERR_TIMEOUT                = 42     # Operation timed out.
    ERR_CANCELED               = 43     # Operation was canceled.
    ERR_PROTOCOL_UNSUPPORTED   = 44     # The protocol is not supported.
    NO_ACCESS                  = 48     # Invalid access for this command.
    ERR_UNSAFE                 = 49     # Invalid access for this command.
    NOT_AVAILABLE              = 64     # Command currently not available.
    NOT_IMPLEMENTED            = 65     # Command not implemented (not yet or not anymore).
    NOT_INITIALIZED            = 67     # Something must first be initialized.
    ERR_NOT_STARTED            = 68     # Something must first be started.
    WRITE_DISABLED             = 80     # Write is disabled for given type.
    ERR_WRITE_NOT_ALLOWED      = 81     # Direct write is not allowed for this type, use command instead.
    ADC_INVALID_CHANNEL        = 96     # Invalid adc input channel selected.
    ERR_EVENT_UNHANDLED        = 112    # The event or command was not handled.
    UNSPECIFIED                = 65535

    @classmethod
    def has_value(cls, value):
        return any(value == item.value for item in cls)


class ProcessType(IntEnum):
    CONTINUE                   = 0
    FINISHED                   = 1
    ABORT_ERROR                = 2


class BroadcastTypes(IntEnum):
    NO_OP                      = 0
    MULTI_SWITCH               = 1
    SET_TIME                   = 2
    BEHAVIOUR_SETTINGS         = 3


class GetPersistenceMode(IntEnum):
    CURRENT = 0
    STORED = 1
    FIRMWARE_DEFAULT = 2

class SetPersistenceMode(IntEnum):
    TEMPORARY = 0
    STORED = 1

class PowerSamplesType(IntEnum):
    SWITCHCRAFT                = 0
    SWITCHCRAFT_NON_TRIGGERED  = 1
    NOW_FILTERED               = 2
    NOW_UNFILTERED             = 3
    SOFT_FUSE                  = 4
    SWITCH                     = 5
    UNSPECIFIED                = 255

class CommandSourceType(IntEnum):
    ENUM                       = 0
    BEHAVIOUR                  = 1
    BROADCAST                  = 3
    UNSPECIFIED                = 255

class CommandSourceId(IntEnum):
    NONE                       = 0
    INTERNAL                   = 2
    UART                       = 3
    CONNECTION                 = 4
    SWITCHCRAFT                = 5
    TAP_TO_TOGGLE              = 6
    UNSPECIFIED                = 255

class MicroappOpcode(IntEnum):
    UPLOAD                     = 1
    VALIDATE                   = 2
    ENABLE                     = 3
    DISABLE                    = 4
    REQUEST                    = 5
    UNSPECIFIED                = 255

class SwitchValSpecial(IntEnum):
    TOGGLE                     = 253 # Switch OFF when currently on, switch to SMART_ON when currently off.
    BEHAVIOUR                  = 254 # Switch to the value according to behaviour rules.
    SMART_ON                   = 255 # Switch on, the value will be determined by behaviour rules.



from crownstone_core.util.DataStepper import DataStepper
from crownstone_core.protocol.BluenetTypes import CommandSourceType, CommandSourceId
from crownstone_core.Exceptions import CrownstoneError, CrownstoneException

class CommandSourcePacket:
	def __init__(self, data):
		self.sourceType = CommandSourceType.UNSPECIFIED  # Bits 5-7
		self.viaMesh = False  # Bit 0
		self.sourceId = CommandSourceId.UNSPECIFIED  # Uint 8
		self.load(data)

	def load(self, data):
		"""
		Parses data buffer to set member variables.

		data : list of bytes

		Raises exception when parsing fails.
		"""
		streamBuf = DataStepper(data)
		byte1 = streamBuf.getUInt8()
		sourceTypeval = (byte1 >> 5) & 7
		self.sourceType = CommandSourceType(sourceTypeval) # Throws exception of value is not in enum
		self.viaMesh = (byte1 & (1 << 0)) != 0
		self.sourceId = streamBuf.getUInt8()

	@staticmethod
	def size():
		return 1 + 1

	def toString(self):
		msg = "CommandSourcePacket("
		msg += "type=" + str(self.sourceType)
		msg += " viaMesh=" + str(self.viaMesh)
		msg += " sourceId=" + str(self.sourceId)
		msg += ")"
		return msg

	def __str__(self):
		return self.toString()


from crownstone_core.protocol.BluenetTypes import ResultValue, ControlType
from crownstone_core.util.DataStepper import DataStepper


class ResultPacket:
    
    def __init__(self, data) :
        self.commandType = ControlType.UNSPECIFIED
        self.protocol = 0
        self.size = 0
        self.commandTypeUInt16 = 0
        self.resultCode = 0
        self.valid = True

        self.payload = []
        self.load(data)
    

    def load(self, data) :
        minSize = 7

        if len(data) >= minSize:
            payload = DataStepper(data)

            self.protocol = payload.getUInt8()
            self.commandTypeUInt16 = payload.getUInt16()
            resultNumber  = payload.getUInt16()

            if ControlType.has_value(self.commandTypeUInt16) and ResultValue.has_value(resultNumber):
                self.commandType = ControlType(self.commandTypeUInt16)
                self.resultCode  = ResultValue(resultNumber)
                self.size        = payload.getUInt16()

                totalSize = minSize + self.size
                if len(data) >= totalSize:
                    if self.size == 0:
                        return

                    for i in range(minSize, totalSize):
                        self.payload.append(data[i])
                else:
                    self.valid = False
            else:
                self.valid = False
        else:
            self.valid = False


class Topics:
    
    advertisement = "advertisement" # data is dictionary {
    #  name: string
    #  rssi: int
    #  address: string   # mac address
    #  serviceUUID: string
    #  serviceData: {
    #    opCode:                       int
    #    dataType:                     int
    #    stateOfExternalCrownstone:    int    # adv contains state of external crownstone
    #    hasError:                     bool   # this crownstone has an error
    #    setupMode:                    bool   # is in setup mode
    #    id:                           int    # crownstone id (0-255)
    #    switchState:                  int    # raw switch state (uint8)
    #    flagsBitmask:                 int
    #    temperature:                  int    # chip temp
    #    powerFactor:                  int    # factor between real and apparent
    #    powerUsageReal:               int    # usage in watts (W)
    #    powerUsageApparent:           int    # usage in VA
    #    accumulatedEnergy:            int
    #    timestamp:                    int    # time on Crownstone seconds since epoch with locale correction
    #    dimmerReady:                  bool   # dimming is available for use (it is not in the first 60 seconds after boot)
    #    dimmingAllowed:               bool   # this Crownstone can dim
    #    switchLocked:                 bool   # this Crownstone is switch-locked
    #    errorMode:                    bool   # advertisement type errorMode : the errors JSON is valid. This alternates with normal advertisements
    #    errors: {
    #        overCurrent:              bool
    #        overCurrentDimmer:        bool
    #        temperatureChip:          bool
    #        temperatureDimmer:        bool
    #        dimmerOnFailure:          bool
    #        dimmerOffFailure:         bool
    #        bitMask:                  int
    #    }
    #    uniqueElement:                int    # something that identifies this advertisement uniquely. Can be used to skip duplicate payloads
    #    timeIsSet:                    bool   # this crownstone knows what time it is
    #  }

    newDataAvailable = "newDataAvailable"  # data is dictionary {
    #   id:                           int    # crownstone id (0-255)
    #   setupMode:                    bool   # is in setup mode
    #   switchState:                  int    # raw switch state (uint8)
    #   temperature:                  int    # chip temp in Celcius
    #   powerFactor:                  int    # factor between real and apparent
    #   powerUsageReal:               int    # power usage in watts (W)
    #   powerUsageApparent:           int    # power usage in VA
    #   accumulatedEnergy:            int
    #   timestamp:                    int    # time on Crownstone seconds since epoch with locale correction
    #   dimmerReady:                  bool   # dimming is available for use (it is not in the first 60 seconds after boot)
    #   dimmingAllowed:               bool   # this Crownstone can dim
    #   switchLocked:                 bool   # this Crownstone is switch-locked
    #   hasError:                     bool   # this crownstone has an error, if the crownstone has an error, the errors: {} dict is only valid if errorMode: true. This boolean is always valid.
    #   errorMode:                    bool   # summary type errorMode : the errors JSON is valid. This alternates with normal advertisements
    #   errors: {
    #       overCurrent:              bool
    #       overCurrentDimmer:        bool
    #       temperatureChip:          bool
    #       temperatureDimmer:        bool
    #       dimmerOnFailure:          bool
    #       dimmerOffFailure:         bool
    #       bitMask:                  int
    #   }
    #   timeIsSet:                    bool   # this crownstone knows what time it is
    # }




from crownstone_core.Constants import UserLevel

from crownstone_core.Exceptions import CrownstoneBleException, EncryptionError
from crownstone_core.protocol.BlePackets import SUPPORTED_PROTOCOL_VERSION
from crownstone_core.util.Conversion import Conversion
from crownstone_core.util.EncryptionHandler import SESSION_KEY_LENGTH, SESSION_DATA_LENGTH


class EncryptionSettings:
    
    def __init__(self):
        self.adminKey = None
        self.memberKey = None
        self.basicKey = None
        self.setupKey = None
        self.serviceDataKey = None
        self.localizationKey = None
        self.meshApplicationKey = None
        self.meshNetworkKey = None

        self.referenceId = None
        self.sessionNonce = None
        self.crownstoneProtocolVersion = None
        self.validationKey = None
        self.initializedKeys = False
        self.temporaryDisable = False
        self.userLevel = UserLevel.unknown


    def loadKeys(self, adminKey, memberKey, basicKey, serviceDataKey, localizationKey, meshApplicationKey, meshNetworkKey, referenceId):
        self.adminKey  = Conversion.ascii_or_hex_string_to_16_byte_array(adminKey)
        self.memberKey = Conversion.ascii_or_hex_string_to_16_byte_array(memberKey)
        self.basicKey  = Conversion.ascii_or_hex_string_to_16_byte_array(basicKey)
        self.serviceDataKey  = Conversion.ascii_or_hex_string_to_16_byte_array(serviceDataKey)
        self.localizationKey  = Conversion.ascii_or_hex_string_to_16_byte_array(localizationKey)
        self.meshApplicationKey = Conversion.ascii_or_hex_string_to_16_byte_array(meshApplicationKey)
        self.meshNetworkKey  = Conversion.ascii_or_hex_string_to_16_byte_array(meshNetworkKey)

        self.referenceId = referenceId
        
        self.initializedKeys = True
        self.determineUserLevel()
        

    def determineUserLevel(self):
        if   len(self.adminKey)  == 16:
            self.userLevel = UserLevel.admin
        elif len(self.memberKey) == 16:
            self.userLevel = UserLevel.member
        elif len(self.basicKey)  == 16:
            self.userLevel = UserLevel.basic
        else:
            self.userLevel = UserLevel.unknown
            self.initializedKeys = False


    def invalidateSessionNonce(self):
        self.sessionNonce = None
        self.validationKey = None
        self.crownstoneProtocolVersion = 0


    def setSessionNonce(self, sessionNonce):
        if len(sessionNonce) != SESSION_DATA_LENGTH:
            raise CrownstoneBleException(EncryptionError.INVALID_SESSION_DATA, "Invalid Session Data")
        self.sessionNonce = sessionNonce

    def setValidationKey(self, validationKey):
        if len(validationKey) != SESSION_KEY_LENGTH:
            raise CrownstoneBleException(EncryptionError.INVALID_SESSION_DATA, "Invalid Session Data")
        self.validationKey = validationKey

    def setCrownstoneProtocolVersion(self, protocolVersion):
        self.crownstoneProtocolVersion = protocolVersion
        if protocolVersion != SUPPORTED_PROTOCOL_VERSION:
            print("Warning: Crownstone has protocol version", protocolVersion, "while the library uses", SUPPORTED_PROTOCOL_VERSION, ". Check if you have the correct version of the library for your firmware version.")



    def loadSetupKey(self, setupKey):
        self.setupKey  = setupKey
        self.userLevel = UserLevel.setup


    def exitSetup(self):
        self.setupKey = None
        self.determineUserLevel()


    def disableEncryptionTemporarily(self):
        self.temporaryDisable = True


    def restoreEncryption(self):
        self.temporaryDisable = False


    def isTemporarilyDisabled(self):
        return self.temporaryDisable


    def isEncryptionEnabled(self):
        if self.temporaryDisable:
            return False
        return True









from crownstone_core.util.Conversion import Conversion


def fletcher32_uint8Arr(uint8Array):
    uint16Array = Conversion.uint8_array_to_uint16_array(uint8Array)
    return fletcher32_uint16Arr(uint16Array)

def fletcher32_uint16Arr(uint16Array):
    uint32_max = 4294967295
    c0 = 0
    c1 = 0
    iterations = int(len(uint16Array) / 360)
    length = len(uint16Array)
    index = 0
    for i in range(iterations + 1):
        blockLength = min(360, length)
        for j in range(blockLength):
            c0 = (c0 + uint16Array[index]) % uint32_max
            c1 = c1 + c0
            index += 1

        c0 = c0 % 65535
        c1 = c1 % 65535
        length -= 360

    return ((c1 << 16) % uint32_max) | c0



from enum import IntEnum

from crownstone_core.util.Conversion import Conversion
from crownstone_core.util.DataStepper import DataStepper


class BehaviourTimeType(IntEnum):
    afterMidnight = 0
    afterSunrise  = 1
    afterSunset   = 2

    @classmethod
    def has_value(cls, value):
        return any(value == item.value for item in cls)




class BehaviourTimeContainer:
    def __init__(self, fromTime, untilTime):
        self.fromTime  = fromTime
        self.untilTime = untilTime

class BehaviourTime:

    def __init__(self):
        self.timeType = None
        self.offset = 0
        self.valid = True

    def fromTime(self, hours, minutes):
        self.timeType = BehaviourTimeType.afterMidnight
        self.offset = 3600 * hours + 60 * minutes
        return self

    def fromType(self, timeType, offsetSeconds=0):
        self.timeType = timeType
        self.offset = offsetSeconds
        return self

    def fromData(self, data):
        if len(data) != 5:
            self.valid = False
            return self

        payload = DataStepper(data)

        firstByte = payload.getUInt8()
        if not BehaviourTimeType.has_value(firstByte):
            self.valid = False
            return self

        self.timeType = BehaviourTimeType(firstByte)
        self.offset = payload.getInt32()
        self.valid = True

        return self

    def getPacket(self):
        arr = []

        arr.append(self.timeType.value)
        arr += Conversion.int32_to_uint8_array(self.offset)

        return arr

    def getDictionary(self):
        returnDict = {}

        if self.timeType == BehaviourTimeType.afterSunset:
            returnDict["type"] = "SUNSET"
            returnDict["offsetMinutes"] = self.offset / 60
        elif self.timeType == BehaviourTimeType.afterSunrise:
            returnDict["type"] = "SUNRISE"
            returnDict["offsetMinutes"] = self.offset / 60
        else:
            returnDict["type"] = "CLOCK"
            returnDict["data"] = {"hours": (self.offset - self.offset % 3600) / 3600,
                                  "minutes": (self.offset % 3600) / 60}

        return returnDict




CROWNSTONE_PLUG_ADVERTISEMENT_SERVICE_UUID = "C001"
CROWNSTONE_BUILTIN_ADVERTISEMENT_SERVICE_UUID = "C002"
CROWNSTONE_GUIDESTONE_ADVERTISEMENT_SERVICE_UUID  = "C003"
DFU_ADVERTISEMENT_SERVICE_UUID = "00001530-1212-EFDE-1523-785FEABCD123"

class CSServices:
	DeviceInformation         = "180a"
	CrownstoneService         = "24f00000-7d10-4805-bfc1-7663a01c3bff"
	SetupService              = "24f10000-7d10-4805-bfc1-7663a01c3bff"
	GeneralService            = "24f20000-7d10-4805-bfc1-7663a01c3bff"
	PowerService              = "24f30000-7d10-4805-bfc1-7663a01c3bff"
	IndoorLocalizationService = "24f40000-7d10-4805-bfc1-7663a01c3bff"
	ScheduleService           = "24f50000-7d10-4805-bfc1-7663a01c3bff"
	MeshService               = "0000fee4-0000-1000-8000-00805f9b34fb"


class DFUServices:
	DFU = DFU_ADVERTISEMENT_SERVICE_UUID




from crownstone_core.packets.behaviour.PresenceDescription import BehaviourPresenceType, BehaviourPresence
from crownstone_core.packets.behaviour.SwitchBehaviour import SwitchBehaviour


class ExtendedSwitchBehaviour(SwitchBehaviour):
    def __init__(self, *args):
        super().__init__(*args)
        self.endCondition = None

    def setNoEndCondition(self):
        self.endCondition = None
        return self

    def setEndConditionWhilePeopleInSphere(self):
        self.endCondition = BehaviourPresence().setSpherePresence(BehaviourPresenceType.someoneInSphere)
        return self

    def setEndConditionWhilePeopleInLocation(self, locationId):
        self.endCondition = BehaviourPresence().setLocationPresence(BehaviourPresenceType.somoneInLocation,
                                                                    [locationId])
        return self

    def getPacket(self):
        arr = super().getPacket()

        if self.endCondition is not None:
            arr += self.endCondition.getPacket()
        else:
            anypresence = BehaviourPresence()
            arr += anypresence.getPacket()

        return arr
